/*
 *  TraceAlloc.c by Davide Libenzi (Trace allocator for iPhone/OSX)
 *  Copyright (C) 2008  Davide Libenzi
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Davide Libenzi <davidel@xmailserver.org>
 *
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <dlfcn.h>
#include <malloc/malloc.h>
#import <CoreFoundation/CoreFoundation.h>
#import <Foundation/Foundation.h>
#include "DblList.h"
#include "TraceAlloc.h"

#define TA_CTX_MAGIC "TRACEATR"
#define TA_CTX_MAGIC_LEN (sizeof(TA_CTX_MAGIC) - 1)
#define TA_MAX_TRACE 16
#define TA_MAX_STACK_DISP (1024 * 64)
#define TA_BLOCKHDR_EXTRA sizeof(struct TracedBlockHdr)

/*
 * It happened many times with half broken GCC releases, that the return
 * value of __builtin_frame_address() and __builtin_return_address() macros
 * have been bogus. And trying to continue to hike up the chain could have
 * lead to SEGV. The macro below tries to protect us from bogus values returned
 * by __builtin_frame_address(), and once we hit one of them, we stop the
 * unwind loop. In theory __builtin_frame_address() should return zero at the
 * end of the trace, and we cover this one too.
 */
#define TA_FETCH_FRAME(res, tos, n)				\
	do {							\
		long disp;					\
		char *cur;					\
		cur = (char *) __builtin_frame_address(n);	\
		disp = (long) (cur - tos);			\
		if (cur == NULL || disp < -TA_MAX_STACK_DISP || \
		    disp > TA_MAX_STACK_DISP)			\
			res = 0;				\
		else {						\
			res = 1;				\
			tos = cur;				\
		}						\
	} while (0)
#define TA_BRA(n) __builtin_extract_return_addr(__builtin_return_address(n))
#define TA_FETCH_ROUND(r)						\
	do {								\
		TA_FETCH_FRAME(res, tos, r + 2);			\
		if (!res || n < (r + 1) || (trace[r] = TA_BRA(r + 2)) == NULL) \
			return r;					\
	} while (0)

struct TracedBlockHdr {
	struct DblListHead Lnk;
	unsigned long mark;
	unsigned long size;
	void *callTrace[TA_MAX_TRACE];
};

struct TraceAllocatorContext {
	char magic[TA_CTX_MAGIC_LEN];
	CFAllocatorRef origATor;
	int refCount;
	unsigned long mark;
	CFAllocatorContext ctx;
	NSLock *lock;
	struct DblListHead blksHead;
	unsigned long allocSize;
	unsigned long numBlocks;
};

static int TAFillCallTrace(void **trace, int n)
{
	int res;
	char *tos;

	/*
	 * You may say: "WTF are you unrolling the loop for?"
	 * Well, GCC's __builtin_* macros wants const parameters, so
	 * we need to open code the loop.
	 */
	tos = (char *) __builtin_frame_address(0);
	TA_FETCH_FRAME(res, tos, 1);
	if (!res || TA_BRA(1) == NULL)
		return 0;
	TA_FETCH_ROUND(0);
	TA_FETCH_ROUND(1);
	TA_FETCH_ROUND(2);
	TA_FETCH_ROUND(3);
	TA_FETCH_ROUND(4);
	TA_FETCH_ROUND(5);
	TA_FETCH_ROUND(6);
	TA_FETCH_ROUND(7);
	TA_FETCH_ROUND(8);
	TA_FETCH_ROUND(9);
	TA_FETCH_ROUND(10);
	TA_FETCH_ROUND(11);
	TA_FETCH_ROUND(12);
	TA_FETCH_ROUND(13);
	TA_FETCH_ROUND(14);
	TA_FETCH_ROUND(15);
	TA_FETCH_ROUND(16);
	TA_FETCH_ROUND(17);
	TA_FETCH_ROUND(18);
	TA_FETCH_ROUND(19);
	TA_FETCH_ROUND(20);
	TA_FETCH_ROUND(21);
	TA_FETCH_ROUND(22);
	TA_FETCH_ROUND(23);
	TA_FETCH_ROUND(24);
	TA_FETCH_ROUND(25);
	TA_FETCH_ROUND(26);
	TA_FETCH_ROUND(27);
	TA_FETCH_ROUND(28);
	TA_FETCH_ROUND(29);
	TA_FETCH_ROUND(30);
	TA_FETCH_ROUND(31);
	/*
	 * OK, should be enough to capture the whole trace ...
	 */

	return 32;
}

static void TAGetCallTrace(void **trace, int n)
{
	int m;

	m = TAFillCallTrace(trace, n);
	if (m < n)
		memset(&trace[m], 0, (n - m) * sizeof(void *));
}

/*
 * We have to use functions that do not alloc any memory, inside
 * the allocator dumping functions. We're holding its lock, so if
 * another memory allocation comes in, we deadlock.
 */
static int TAGetSymInfo(void *addr, char *buf, int n, int flags)
{
	const char *name;
	Dl_info symInfo;

	if (dladdr(addr, &symInfo) == 0)
		return -1;
	name = (const char *) symInfo.dli_fname;
	if ((flags & TAF_FULL_NAME) == 0) {
		if ((name = strrchr((const char *) symInfo.dli_fname, '/')) == NULL)
			name = (const char *) symInfo.dli_fname;
		else
			name++;
	}
	if (flags & TAF_BRIEF_SYMBOL)
		return snprintf(buf, n, "`%s@%p\t%p", name, addr, symInfo.dli_fbase);

	return snprintf(buf, n, "`%s@%p\t%p\t%s\t%p (+0x%04x)",
			name, addr, symInfo.dli_fbase,
			symInfo.dli_sname, symInfo.dli_saddr,
			(int) (((char *) addr) - ((char *) symInfo.dli_saddr)));
}

static void TADumpBlockInfo(struct TraceAllocatorContext *tac,
			    struct TracedBlockHdr *bhdr, char *buf, int size, int fd,
			    int flags)
{
	int i, n;

	n = snprintf(buf, size,
		     "Block @ %p (%lu bytes)\n"
		     "--> Trace:\n",
		     (char *) bhdr + TA_BLOCKHDR_EXTRA, bhdr->size);
	write(fd, buf, n);
	for (i = 0; i < TA_MAX_TRACE && bhdr->callTrace[i] != NULL; i++) {
		if ((n = TAGetSymInfo(bhdr->callTrace[i], buf, size, flags)) < 0)
			n = snprintf(buf, size, "Unknown symbol @ %p\n",
				     bhdr->callTrace[i]);
		write(fd, "\t", 1);
		write(fd, buf, n);
		write(fd, "\n", 1);
	}
	write(fd, "\n", 1);
}

static void TAAddBlock(struct TraceAllocatorContext *tac, unsigned long size,
		       struct TracedBlockHdr *bhdr)
{
	bhdr->size = size;
	TAGetCallTrace(bhdr->callTrace, TA_MAX_TRACE);
	[tac->lock lock];
	DBLL_LIST_ADDT(&bhdr->Lnk, &tac->blksHead);
	tac->allocSize += bhdr->size;
	tac->numBlocks++;
	bhdr->mark = tac->mark;
	[tac->lock unlock];
}

static void TARemoveBlock(struct TraceAllocatorContext *tac,
			  struct TracedBlockHdr *bhdr)
{
	[tac->lock lock];
	DBLL_LIST_DEL(&bhdr->Lnk);
	tac->allocSize -= bhdr->size;
	tac->numBlocks--;
	[tac->lock unlock];
}

static const void *TAAllocatorRetainCallBack(const void *info)
{
	struct TraceAllocatorContext *tac = (struct TraceAllocatorContext *) info;

	tac->refCount++;

	return (void *) info;
}

static void TAAllocatorReleaseCallBack(const void *info)
{
	struct TraceAllocatorContext *tac = (struct TraceAllocatorContext *) info;

	tac->refCount--;
	if (tac->refCount == 0) {

		[tac->lock release];
		free(tac);
	}
}

static CFStringRef TAAllocatorCopyDescriptionCallBack(const void *info)
{
	struct TraceAllocatorContext *tac = (struct TraceAllocatorContext *) info;
	NSString *origDesc, *atorDesc;
	CFAllocatorContext origCtx;

	CFAllocatorGetContext(tac->origATor, &origCtx);
	origDesc = (NSString *) (*origCtx.copyDescription)(origCtx.info);
	atorDesc = [[NSString alloc] initWithFormat: @"Hooked Allocator (%@)", origDesc];
	[origDesc release];

	return (CFStringRef) atorDesc;
}

static void *TAAllocatorAllocateCallBack(CFIndex allocSize, CFOptionFlags hint, void *info)
{
	struct TraceAllocatorContext *tac = (struct TraceAllocatorContext *) info;
	void *ptr;
	struct TracedBlockHdr *bhdr;

	if ((ptr = CFAllocatorAllocate(tac->origATor, allocSize + TA_BLOCKHDR_EXTRA,
				       hint)) != NULL) {
		TAAddBlock(tac, allocSize, (struct TracedBlockHdr *) ptr);
		ptr = (char *) ptr + TA_BLOCKHDR_EXTRA;
	}

	return ptr;
}

static void *TAAllocatorReallocateCallBack(void *ptr, CFIndex newsize, CFOptionFlags hint,
					   void *info)
{
	struct TraceAllocatorContext *tac = (struct TraceAllocatorContext *) info;
	void *nptr;

	if (ptr != NULL) {
		ptr = (char *) ptr - TA_BLOCKHDR_EXTRA;
		TARemoveBlock(tac, (struct TracedBlockHdr *) ptr);
	}
	if ((nptr = CFAllocatorReallocate(tac->origATor, ptr, newsize + TA_BLOCKHDR_EXTRA,
					  hint)) != NULL) {
		TAAddBlock(tac, newsize, (struct TracedBlockHdr *) nptr);
		nptr = (char *) nptr + TA_BLOCKHDR_EXTRA;
	}

	return nptr;
}

static void TAAllocatorDeallocateCallBack(void *ptr, void *info)
{
	struct TraceAllocatorContext *tac = (struct TraceAllocatorContext *) info;
	struct TracedBlockHdr *bhdr;

	if (ptr != NULL) {
		ptr = (char *) ptr - TA_BLOCKHDR_EXTRA;
		TARemoveBlock(tac, (struct TracedBlockHdr *) ptr);
		CFAllocatorDeallocate(tac->origATor, ptr);
	}
}

static CFIndex TAAllocatorPreferredSizeCallBack(CFIndex size, CFOptionFlags hint,
						void *info)
{
	struct TraceAllocatorContext *tac = (struct TraceAllocatorContext *) info;

	return CFAllocatorGetPreferredSizeForSize(tac->origATor, size, hint) +
		TA_BLOCKHDR_EXTRA;
}

static void TADumpAllocatedBlocks(struct TraceAllocatorContext *tac, unsigned long mark,
				  int fd, int flags)
{
	int n;
	struct TracedBlockHdr *bhdr;
	struct DblListHead *pos;
	char buf[512];

	[tac->lock lock];
	n = snprintf(buf, sizeof(buf), "Allocated size: %lu bytes\n\n", tac->allocSize);
	write(fd, buf, n);
	DBLL_LIST_FOR_EACH(pos, &tac->blksHead) {
		bhdr = DBLL_LIST_ENTRY(pos, struct TracedBlockHdr, Lnk);
		if (bhdr->mark == mark)
			TADumpBlockInfo(tac, bhdr, buf, sizeof(buf), fd, flags);
	}
	[tac->lock unlock];
}

static struct TraceAllocatorContext *TAGetContext(CFAllocatorRef ator)
{
	struct TraceAllocatorContext *tac;
	CFAllocatorContext ctx;

	CFAllocatorGetContext(ator, &ctx);
	tac = (struct TraceAllocatorContext *) ctx.info;

	return memcmp(tac->magic, TA_CTX_MAGIC, TA_CTX_MAGIC_LEN) == 0 ? tac: NULL;
}

int TAGetStatus(CFAllocatorRef hookATor, struct TraceAllocatorStatus *status)
{
	struct TraceAllocatorContext *tac;

	if ((tac = TAGetContext(hookATor)) == NULL)
		return -1;
	memset(status, 0, sizeof(*status));
	[tac->lock lock];
	status->mark = tac->mark;
	status->allocSize = tac->allocSize;
	status->numBlocks = tac->numBlocks;
	[tac->lock unlock];

	return 0;
}

int TADumpBlocks(CFAllocatorRef hookATor, unsigned long mark, int fd, int flags)
{
	struct TraceAllocatorContext *tac;

	if ((tac = TAGetContext(hookATor)) == NULL)
		return -1;
	TADumpAllocatedBlocks(tac, mark, fd, flags);

	return 0;
}

int TADumpBlocksFile(CFAllocatorRef hookATor, unsigned long mark,
		     const char *path, int flags)
{
	int fd, error = -1;

	NSLog(@"Dumping allocator blocks: mark=%lu file='%s'\n", mark, path);
	if ((fd = open(path, O_WRONLY | O_CREAT | O_TRUNC,
		       0664)) != -1) {
		error = TADumpBlocks(hookATor, mark, fd, flags);
		close(fd);
	} else
		NSLog(@"Unable to create file: file='%s'\n", path);

	return error;
}

unsigned long TAAllocMark(CFAllocatorRef hookATor)
{
	unsigned long mark;
	struct TraceAllocatorContext *tac;

	if ((tac = TAGetContext(hookATor)) == NULL)
		return 0;
	[tac->lock lock];
	mark = ++tac->mark;
	[tac->lock unlock];

	return mark;
}

int TAAllocUnmark(CFAllocatorRef hookATor, int fd, int flags)
{
	unsigned long mark;
	struct TraceAllocatorContext *tac;

	if ((tac = TAGetContext(hookATor)) == NULL)
		return -1;
	[tac->lock lock];
	mark = tac->mark--;
	[tac->lock unlock];
	if (fd != -1)
		TADumpAllocatedBlocks(tac, mark, fd, flags);

	return 0;
}

int TAAllocUnmarkFile(CFAllocatorRef hookATor, const char *path, int flags)
{
	int fd, error = -1;

	NSLog(@"Dumping allocator blocks: file='%s'\n", path);
	if ((fd = open(path, O_WRONLY | O_CREAT | O_TRUNC,
		       0664)) != -1) {
		error = TAAllocUnmark(hookATor, fd, flags);
		close(fd);
	} else
		NSLog(@"Unable to create file: file='%s'\n", path);

	return error;
}

CFAllocatorRef TAAllocatorCreate(CFAllocatorRef origATor)
{
	struct TraceAllocatorContext *tac;
	CFAllocatorRef tracedATor;

	if ((tac = (struct TraceAllocatorContext *)
	     malloc(sizeof(struct TraceAllocatorContext))) == NULL)
		return NULL;
	memset(tac, 0, sizeof(*tac));
	memcpy(tac->magic, TA_CTX_MAGIC, TA_CTX_MAGIC_LEN);
	tac->origATor = origATor;
	tac->refCount = 1;
	DBLL_INIT_LIST_HEAD(&tac->blksHead);
	tac->ctx.version = 1;
	tac->ctx.info = tac;
	tac->ctx.retain = TAAllocatorRetainCallBack;
	tac->ctx.release = TAAllocatorReleaseCallBack;
	tac->ctx.copyDescription = TAAllocatorCopyDescriptionCallBack;
	tac->ctx.allocate = TAAllocatorAllocateCallBack;
	tac->ctx.reallocate = TAAllocatorReallocateCallBack;
	tac->ctx.deallocate = TAAllocatorDeallocateCallBack;
	tac->ctx.preferredSize = TAAllocatorPreferredSizeCallBack;
	if ((tracedATor = CFAllocatorCreate(origATor, &tac->ctx)) == NULL) {
		NSLog(@"Unable to create traced allocator\n");
		free(tac);
		return NULL;
	}
	tac->lock = [[NSLock alloc] init];

	return tracedATor;
}

int TATraceDefaultAllocator(void)
{
	CFAllocatorRef tracedATor;

	NSLog(@"Setting tracing allocator\n");
	if ((tracedATor = TAAllocatorCreate(CFAllocatorGetDefault())) == NULL)
		return -1;
	TAAllocMark(tracedATor);
	CFAllocatorSetDefault(tracedATor);
	NSLog(@"Tracing allocator installed: ator=%p\n", tracedATor);

	return 0;
}

